''' Utils helping with hex<->string conversion and stripping '''


def strip_0x04(p: bytes):
    ''' Strip the 0x04 off the starting of a byte sequence.'''
    if len(p) == 65 and p[0] == 4:
        return p[1:]
    else:
        return p


def remove_0x(address: str) -> str:
    '''
    Remove the 0x if any. Returns the string without 0x

    Parameters
    ----------
    address : str
        Address string, like 0xabc...

    Returns
    -------
    str
        Address string without prefix "0x"
    '''

    if address.startswith("0x") or address.startswith("0X"):
        return address[2:]
    else:
        return address


def is_uncompressed_public_key(key_bytes: bytes) -> bool:
    '''
    Check if bytes is the uncompressed public key.

    Parameters
    ----------
    address : bytes
        Address in bytes. Should be 65 bytes.

    Returns
    -------
    bool
        True/False

    Raises
    ------
    ValueError
        If address isn't 65 bytes.
    ValueError
        If address doesn't begin with 04 as first byte.
    '''
    if len(key_bytes) != 65:
        raise ValueError('Length should be 65 bytes.')

    if key_bytes[0] != 4:
        raise ValueError('Should begin with 04 as first byte.')

    return True



'''
Blake2b

Blake2b hash function.
'''
import hashlib  # python3 lib/hashlib
from typing import List, Tuple


def blake2b256(list_of_bytes: List[bytes]) -> Tuple[bytes, int]:
    '''
    Computes a hash in black2b flavor, the output is 256 bits / 32 bytes.

    Parameters
    ----------
    list_of_bytes : List[bytes]
        The list of bytes, waited to be hashed.

    Returns
    -------
    Tuple[bytes, int]
        Hash result in bytes and the length of bytes (32).
    '''

    m = hashlib.blake2b(digest_size=32)
    for item in list_of_bytes:
        m.update(item)

    return m.digest(), m.digest_size



'''
ABI Module.

ABI structure the "Functions" and "Events".

ABI also encode/decode params for functions.

See:
https://github.com/ethereum/wiki/wiki/Ethereum-Contract-ABI

"Function Selector":
sha3("funcName(uint256,address)") -> cut out first 4 bytes.

"Argument Encoding":

Basic:
uint<M> M=8,16,...256
int<M> M=8,16,...256
address
bool
fixed<M>x<N> fixed256x18
bytes<M> bytes32
function 20bytes address + 4 bytes signature.

Fixed length:
<type>[M] Fix sized array. int[10], uint256[33], 

Dynamic length:
bytes
string
<type>[]
'''

# voluptuous is a better library in validating dict.
from voluptuous import Schema, Any, Optional
from typing import List
from typing import Union
import eth_utils
import eth_abi
from .cry import keccak256


MUTABILITY = Schema(Any('pure', 'view', 'constant', 'payable', 'nonpayable'))


FUNC_PARAMETER = Schema({
        "name": str,
        "type": str,
        Optional("components"): list, # if the "type" field is "tuple" or "type[]"
        Optional("internalType"): str
    },
    required=True
)


FUNCTION = Schema({
        "type": "function",
        "name": str,
        Optional("constant"): bool,
        Optional("payable"): bool,
        "stateMutability": MUTABILITY,
        "inputs": [FUNC_PARAMETER],
        "outputs": [FUNC_PARAMETER]
    },
    required=True
)


EVENT_PARAMETER = Schema({
        "name": str,
        "type": str,
        "indexed": bool
    },
    required=True
)


EVENT = Schema({
    "type": "event",
    "name": str,
    Optional("anonymous"): bool,
    "inputs": [EVENT_PARAMETER]
})


def is_dynamic_type(t: str):
    ''' Check if the input type is dynamic '''
    if t == 'bytes' or t == 'string' or t.endswith('[]'):
        return True
    else:
        return False


def dynamic_type_to_topic(t_type:str, value):
    if t_type == 'string':
        return keccak256([value.encode('utf-8')])[0]
    elif t_type == 'bytes':
        return keccak256([value])[0]
    else:
        raise ValueError('complex value type {} is not supported yet, open an issue on Github.'.format(t_type))


def calc_function_selector(abi_json: dict) -> bytes:
    ''' Calculate the function selector (4 bytes) from the abi json '''
    f = FUNCTION(abi_json)
    return eth_utils.function_abi_to_4byte_selector(f)


def calc_event_topic(abi_json: dict) -> bytes:
    ''' Calculate the event log topic (32 bytes) from the abi json'''
    e = EVENT(abi_json)
    return eth_utils.event_abi_to_log_topic(e)


class Coder():
    @staticmethod
    def encode_list(types: List[str], values) -> bytes:
        ''' Encode a sequence of values, into a single bytes '''
        return eth_abi.encode_abi(types, values)

    @staticmethod
    def decode_list(types: List[str], data: bytes) -> List:
        ''' Decode the data, back to a (,,,) tuple '''
        return list(eth_abi.decode_abi(types, data))
    
    @staticmethod
    def encode_single(t: str, value) -> bytes:
        ''' Encode value of type t into single bytes'''
        return Coder.encode_list([t], [value])

    @staticmethod
    def decode_single(t: str, data):
        ''' Decode data of type t back to a single object'''
        return Coder.decode_list([t], data)[0]


class Function():
    def __init__(self, f_definition: dict):
        '''Initialize a function by definition.

        Parameters
        ----------
        f_definition : dict
            See FUNCTION type in this document.
        '''
        self._definition = FUNCTION(f_definition) # Protect.
        self.selector = calc_function_selector(f_definition) # first 4 bytes.
    
    def encode(self, parameters: List, to_hex=False) -> Union[bytes, str]:
        '''Encode the paramters according to the function definition.

        Parameters
        ----------
        parameters : List
            A list of parameters waiting to be encoded.
        to_hex : bool, optional
            If the return should be '0x...' hex string, by default False

        Returns
        -------
        Union[bytes, str]
            Return bytes or '0x...' hex string if needed.
        '''
        my_types = [x['type'] for x in self._definition['inputs']]
        my_bytes = self.selector + Coder.encode_list(my_types, parameters)
        if to_hex:
            return '0x' + my_bytes.hex()
        else:
            return my_bytes
    
    def decode(self, output_data: bytes) -> dict:
        '''Decode function call output data back into human readable results.

        The result is in dual format. Contains both position and named index.
        eg. { '0': 'john', 'name': 'john' }
        '''
        my_types = [x['type'] for x in self._definition['outputs']]
        my_names = [x['name'] for x in self._definition['outputs']]

        result_list = Coder.decode_list(my_types, output_data)

        r = {}
        for idx, name in enumerate(my_names):
            r[str(idx)] = result_list[idx]
            if name:
                r[name] = result_list[idx]
        
        return r


class Event():
    def __init__(self, e_definition: dict):
        '''Initialize an Event with definition.

        Parameters
        ----------
        e_definition : dict
            A dict with style of EVENT.
        '''
        self._definition = EVENT(e_definition)
        self.signature = calc_event_topic(self._definition)

    def encode(self, params: Union[dict, List]) -> List:
        '''Assemble indexed keys into topics.

        Usage
        -----

        Commonly used to filter out logs of concerned topics,
        eg. To filter out VIP180 transfer logs of a certain wallet, certain amount.

        Parameters
        ----------
        params : Union[dict, List]
            A dict/list of indexed param of the given event,
            fill in None to occupy the position,
            if you aren't sure about the value.

            eg. For event:
            
            EventName(address from indexed, address to indexed, uint256 value)

            the params can be: 
            ['0xa32f..ff', '0x1f...ac']
            or:
            {'from': '0xa32f..ff', 'to': '0x1f...ac'}
            or:
            [None, '0x1f...ac']
            or:
            {'from': None, 'to': '0x1f...ac'}

        Returns
        -------
        List
            [description]

        Raises
        ------
        ValueError
            [description]
        '''
        topics = []

        # not anonymous? topic[0] = signature.
        if self._definition.get('anonymous', False) == False:
            topics.append(self.signature)

        indexed_params = [x for x in self._definition['inputs'] if x['indexed']]
        has_no_name_param = any([True for x in indexed_params if not x['name']])

        # Check #1
        if type(params) != list and has_no_name_param:
            raise ValueError('Event definition contains param without a name, use a list of params instead of dict.')

        # Check #2
        if type(params) == list and len(params) != len(indexed_params):
            raise ValueError('Indexed params needs {} length, {} is given.'.format(len(indexed_params), len(params)))

        # Check #3
        if type(params) == dict and len(params.keys()) != len(indexed_params):
            raise ValueError('Indexed params needs {} keys, {} is given.'.format(len(indexed_params), len(params.keys())))

        if type(params) == list:
            for param, definition in zip(params, indexed_params):
                if is_dynamic_type( definition['type'] ):
                    topics.append( dynamic_type_to_topic(definition['type'], param) )
                else:
                    topics.append( Coder.encode_single(definition['type'], param) )

        if type(params) == dict:
            for definition in indexed_params:
                value = params.get(definition['name'], None)
                if value is None:
                    topics.append(value)
                    continue

                if is_dynamic_type( definition['type'] ):
                    topics.append( dynamic_type_to_topic(definition['type'], value) )
                else:
                    topics.append( Coder.encode_single(definition['type'], value) )

        return topics


    def decode(self, data: bytes, topics: List[bytes]):
        ''' Decode "data" according to the "topic"s.

        One output can contain an array of logs[].
        One log contains mainly 3 entries:

        - For a non-indexed parameters event:

            "address": The emitting contract address.
            "topics": [
                "signature of event"
            ]
            "data": "0x..." (contains parameters value)

        - For an indexed parameters event:

            "address": The emitting contract address.
            "topics": [
                "signature of event",
                "indexed param 1",
                "indexed param 2",
                ...
                --> max 3 entries of indexed params.
            ]
            "data": "0x..." (remain un-indexed parameters value)

        If the event is "anonymous" then the signature is not inserted into the "topics" list,
        hence topics[0] is not the signature.
        '''
        if self._definition.get('anonymous', False) == False:
            # if not anonymous, topics[0] is the signature of event.
            # we cut it out, because we already have self.signature
            topics = topics[1:]
        
        _indexed_params_definitions = [x for x in self._definition['inputs'] if x['indexed']]
        _un_indexed_params_definitions = [x for x in self._definition['inputs'] if not x['indexed']]

        if len(_indexed_params_definitions) != len(topics):
            raise Exception('topics count invalid.')
            
        un_indexed_params = Coder.decode_list(
            [x['type'] for x in _un_indexed_params_definitions],
            data
        )

        r = {}
        for idx, each in enumerate(self._definition['inputs']):
            to_be_stored = None
            if each['indexed']:
                topic = topics.pop(0)
                if is_dynamic_type(each['type']):
                    to_be_stored = topic
                else:
                    to_be_stored = Coder.decode_single(each['type'], topic)
            else:
                to_be_stored = un_indexed_params.pop(0)

            r[str(idx)] = to_be_stored

            if each['name']:
                r[each['name']] = to_be_stored

        return r


'''
Keystore Module.

Encrypt, decrypt and verify a keystore.

The "keystore" dict should contain following format:

{
    address: string
    crypto: object
    id: string
    version: number
}

'''
import re
import eth_keyfile

N = 131072  # aka. work_factor
P = 1
R = 8
DK_LEN = 32
SALT_LEN = 16


def encrypt(private_key: bytes, password: bytes) -> dict:
    '''
    Encrypt a private key to a keystore.
    The keystore is a json-style python dict.

    Parameters
    ----------
    private_key : bytes
        A private key in bytes.
    password : bytes
        A password.

    Returns
    -------
    dict
        A keystore
    '''
    return eth_keyfile.create_keyfile_json(private_key, password, 3, "scrypt", N)


def decrypt(keystore: dict, password: bytes) -> bytes:
    '''
    Decrypt a keystore into a private key (bytes).

    Parameters
    ----------
    keystore : dict
        A keystore.
    password : bytes
        A password.

    Returns
    -------
    bytes
        A private key in bytes.
    '''
    return eth_keyfile.decode_keyfile_json(keystore, password)


def _normalize(keystore: dict) -> dict:
    '''
    Normalize the keystore key:value pairs.
    Make each value in lower case.

    Parameters
    ----------
    keystore : dict
        A keystore.

    Returns
    -------
    dict
        A keystore.
    '''
    return keystore


ADDRESS_RE = re.compile('^[0-9a-f]{40}$', re.I)


def _validate(keystore: dict) -> bool:
    '''
    Validate the format of a key store.

    Parameters
    ----------
    keystore : dict
        A keystore.

    Returns
    -------
    bool
        True/False

    Raises
    ------
    ValueError
        If is not in good shape then throw.
    '''
    if keystore.get('version') != 3:
        raise ValueError('unsupported version {}'.format(keystore.version))

    if not ADDRESS_RE.match(keystore.get('address')):
        raise ValueError(
            'invalid address {}, should be 40 characters and alphanumero.'.format(keystore.address))

    if not keystore.get('id'):
        raise ValueError('Need "id" field.')

    if not keystore.get('crypto'):
        raise ValueError('Need "crypto" field.')

    return True


def well_formed(keystore: dict) -> bool:
    '''
    Validate if the keystore is in good shape (roughly).

    Parameters
    ----------
    keystore : dict
        A keystore.

    Returns
    -------
    bool
        True/False
    '''

    return _validate(keystore)



'''
Address Module.

VeChain "public key" and "address" related operations and verifications.
'''

import re
from .keccak import keccak256
from .utils import remove_0x, is_uncompressed_public_key


def public_key_to_address(key_bytes: bytes) -> bytes:
    '''
    Derive an address from a public key
    (uncompressed, starts with 0x04).

    Parameters
    ----------
    key_bytes : bytes
        bytes that represent a public key.

    Returns
    -------
    bytes
        bytes that represents the address.
    '''
    is_uncompressed_public_key(key_bytes)
    # Get rid of the 0x04 (first byte) at the beginning.
    buffer = key_bytes[1:]
    # last 20 bytes from the 32 bytes hash.
    return keccak256([buffer])[0][12:]


def is_address(address: str) -> bool:
    '''
    Check if a text string is valid address.

    Parameters
    ----------
    address : str
        The address string to be checked. Should begin with '0x'.

    Returns
    -------
    bool
        If it is valid address.
    '''

    c = re.compile('^0x[0-9a-f]{40}$', re.I)
    if c.match(address):
        return True
    else:
        return False


def to_checksum_address(address: str) -> str:
    '''
    Turn an address to a checksum address that is compatible with eip-55.

    Parameters
    ----------
    address : str
        The address string. Should begin with '0x'.

    Returns
    -------
    str
        The address that is properly capitalized.

    Raises
    ------
    ValueError
        If the address isn't a valid address itself.
    '''

    if not is_address(address):
        raise ValueError('The address is not valid.')

    body = remove_0x(address)  # remove '0x'.
    body = body.lower()

    h, _ = keccak256([body.encode("ascii")])
    hash = h.hex()

    parts = ['0x']
    for idx, value in enumerate(body):
        if int(hash[idx], 16) >= 8:
            parts.append(value.upper())
        else:
            parts.append(value)

    return ''.join(parts)



import setuptools

long_description = ''
with open("README.md", "r") as fh:
    long_description = fh.read()
assert long_description

setuptools.setup(
    name="thor-devkit",
    version="1.0.6",
    author="laalaguer",
    author_email="laalaguer@gmail.com",
    description="SDK to interact with VeChain Thor public blockchain.",
    keywords="vechain thor blockchain sdk",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    url="https://github.com/laalaguer/thor-devkit.py",
    project_urls={
        'Documentation': 'https://github.com/laalaguer/thor-devkit.py',
        'Source': 'https://github.com/laalaguer/thor-devkit.py',
        'Issue Tracker': 'https://github.com/laalaguer/thor-devkit.py/issues',
    },
    python_requires='>=3.6',
    install_requires=[x.strip() for x in open('requirements.txt')],
    packages=setuptools.find_packages(),
)


'''
Mnemonic Module.

Generate/Validate a words used for mnemonic wallet.

Derive the first private key from words.

Derive the correct seed for BIP32.
'''

from typing import List
from mnemonic import Mnemonic
from bip_utils import Bip32

# BIP-44 specified path notation:
# m / purpose' / coin_type' / account' / change / address_index

# Derive path for the VET:
# m / 44' / 818' / 0' / 0 /<address_index>
VET_PATH = "m/44'/818'/0'/0"


def _get_key_path(base_path: str, index: int = 0) -> str:
    return base_path.rstrip('/') + '/' + str(index)


def _get_vet_key_path(index: int = 0) -> str:
    return _get_key_path(VET_PATH, index)


def generate(strength: int = 128) -> List[str]:
    '''
    Generate BIP39 mnemonic words.

    Parameters
    ----------
    strength : int, optional
         Any of [128, 160, 192, 224, 256], by default 128

    Returns
    -------
    List[str]
        A list of words.

    Raises
    ------
    ValueError
        If the strength is not of correct length.
    '''
    if strength not in [128, 160, 192, 224, 256]:
        raise ValueError(
            'strength should be one of [128, 160, 192, 224, 256].')

    sentence = Mnemonic('english').generate(strength)

    return sentence.split(' ')


def validate(words: List[str]) -> bool:
    '''
    Check if the words form a valid BIP39 mnemonic words.

    Parameters
    ----------
    words : List[str]
        A list of english words.

    Returns
    -------
    bool
        True/False
    '''
    sentence = ' '.join(words)
    return Mnemonic('english').check(sentence)


def derive_seed(words: List[str]) -> bytes:
    '''
    Derive a seed from a word list.

    Parameters
    ----------
    words : List[str]
        A list of english words.

    Returns
    -------
    bytes
        64 bytes
    '''
    sentence = ' '.join(words)
    seed = Mnemonic.to_seed(sentence)  # bytes.
    return seed


def derive_private_key(words: List[str], index: int = 0) -> bytes:
    '''
    Get a private key from the mnemonic wallet,
    default to the 0 index of the deviration. (first key)

    Parameters
    ----------
    words : List[str]
        A list of english words.
    index : int, optional
        The private key index, first private key., by default 0

    Returns
    -------
    bytes
        [description]
    '''
    seed = derive_seed(words)
    bip32_ctx = Bip32.FromSeedAndPath(seed, _get_vet_key_path(index))
    return bip32_ctx.PrivateKey().Raw().ToBytes()
